// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#pragma once

#include <cstdint>
#include <memory>
#include <unordered_map>
#include <vector>

#include "column/column.h"
#include "common/status.h"
#include "exec/parquet/encoding.h"
#include "exec/parquet/level_codec.h"
#include "exec/vectorized/hdfs_scanner.h"
#include "gen_cpp/parquet_types.h"
#include "util/block_compression.h"

namespace starrocks {
class RandomAccessFile;
class BlockCompressionCodec;

} // namespace starrocks

namespace starrocks::parquet {

struct ColumnChunkReaderOptions {
    vectorized::HdfsScanStats* stats = nullptr;
};

class PageReader;

class ColumnChunkReader {
public:
    ColumnChunkReader(level_t max_def_level, level_t max_rep_level, int32_t type_length,
                      const tparquet::ColumnChunk* column_chunk, RandomAccessFile* file,
                      const ColumnChunkReaderOptions& opts);
    ~ColumnChunkReader();

    Status init();

    Status next_page();

    uint32_t num_values() const { return _num_values; }

    // Try to decode n definition levels into 'levels'
    // return number of decoded levels.
    // If the returned value is less than input n, this means current page don't have
    // enough levels.
    // User should call next_page() to get more levels
    size_t decode_def_levels(size_t n, level_t* levels) {
        DCHECK_GT(_max_def_level, 0);
        return _def_level_decoder.decode_batch(n, levels);
    }

    LevelDecoder& def_level_decoder() { return _def_level_decoder; }
    LevelDecoder& rep_level_decoder() { return _rep_level_decoder; }

    size_t decode_rep_levels(size_t n, level_t* levels) {
        DCHECK_GT(_max_rep_level, 0);
        return _rep_level_decoder.decode_batch(n, levels);
    }

    Status decode_values(size_t n, const uint8_t* is_nulls, ColumnContentType content_type, vectorized::Column* dst) {
        size_t idx = 0;
        while (idx < n) {
            bool is_null = is_nulls[idx++];
            size_t run = 1;
            while (idx < n && is_nulls[idx] == is_null) {
                idx++;
                run++;
            }
            if (is_null) {
                dst->append_nulls(run);
            } else {
                RETURN_IF_ERROR(_cur_decoder->next_batch(run, content_type, dst));
            }
        }
        return Status::OK();
    }

    Status decode_values(size_t n, ColumnContentType content_type, vectorized::Column* dst) {
        return _cur_decoder->next_batch(n, content_type, dst);
    }

    const tparquet::ColumnMetaData& metadata() const { return _chunk_metadata->meta_data; }

    Status get_dict_values(vectorized::Column* column) { return _cur_decoder->get_dict_values(column); }

    Status get_dict_values(const std::vector<int32_t>& dict_codes, vectorized::Column* column) {
        return _cur_decoder->get_dict_values(dict_codes, column);
    }

    Status get_dict_codes(const std::vector<Slice>& dict_values, std::vector<int32_t>* dict_codes) {
        return _cur_decoder->get_dict_codes(dict_values, dict_codes);
    }

private:
    Status _parse_page_header();
    Status _parse_page_data();

    Status _try_load_dictionary();
    Status _read_and_decompress_page_data();
    Status _parse_data_page();
    Status _parse_dict_page();

    void _reserve_uncompress_buf(size_t size);
    Status _read_and_decompress_page_data(uint32_t compressed_size, uint32_t uncompressed_size, bool is_compressed);

private:
    enum PageParseState {
        INITIALIZED,
        PAGE_HEADER_PARSED,
        PAGE_LEVELS_PARSED,
        PAGE_DATA_PARSED,
    };

    level_t _max_def_level = 0;
    level_t _max_rep_level = 0;
    int32_t _type_length = 0;
    const tparquet::ColumnChunk* _chunk_metadata = nullptr;
    ColumnChunkReaderOptions _opts;
    std::unique_ptr<RandomAccessFile> _file;

    std::unique_ptr<PageReader> _page_reader;

    const BlockCompressionCodec* _compress_codec = nullptr;

    LevelDecoder _def_level_decoder;
    LevelDecoder _rep_level_decoder;

    size_t _num_values = 0;

    std::unique_ptr<uint8_t[]> _uncompressed_buf;
    size_t _uncompressed_buf_capacity = 0;

    PageParseState _page_parse_state = INITIALIZED;
    Slice _data;

    Decoder* _cur_decoder = nullptr;
    std::unordered_map<int, std::unique_ptr<Decoder>> _decoders;
};

} // namespace starrocks::parquet
