// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#include "formats/csv/nullable_converter.h"

#include "column/nullable_column.h"
#include "common/logging.h"

namespace starrocks::vectorized::csv {

Status NullableConverter::write_string(OutputStream* os, const Column& column, size_t row_num,
                                       const Options& options) const {
    auto nullable_column = down_cast<const NullableColumn*>(&column);
    auto data_column = nullable_column->data_column().get();
    auto null_column = nullable_column->null_column().get();
    if (null_column->get_data()[row_num] != 0) {
        return os->write("\\N");
    } else {
        return _base_converter->write_string(os, *data_column, row_num, options);
    }
}

Status NullableConverter::write_quoted_string(OutputStream* os, const Column& column, size_t row_num,
                                              const Options& options) const {
    auto nullable_column = down_cast<const NullableColumn*>(&column);
    auto data_column = nullable_column->data_column().get();
    auto null_column = nullable_column->null_column().get();
    if (null_column->get_data()[row_num] != 0) {
        return os->write("null");
    } else {
        return _base_converter->write_quoted_string(os, *data_column, row_num, options);
    }
}

bool NullableConverter::read_string(Column* column, Slice s, const Options& options) const {
    auto* nullable = down_cast<NullableColumn*>(column);
    auto* data = nullable->data_column().get();

    if (s == "\\N") {
        return nullable->append_nulls(1);
    } else if (_base_converter->read_string(data, s, options)) {
        nullable->null_column()->append(0);
        return true;
    } else if (options.invalid_field_as_null) {
        return nullable->append_nulls(1);
    } else {
        return false;
    }
}

bool NullableConverter::read_quoted_string(Column* column, Slice s, const Options& options) const {
    auto* nullable = down_cast<NullableColumn*>(column);
    auto* data = nullable->data_column().get();

    if (s == "null") {
        return nullable->append_nulls(1);
    } else if (_base_converter->read_quoted_string(data, s, options)) {
        nullable->null_column()->append(0);
        return true;
    } else if (options.invalid_field_as_null) {
        return nullable->append_nulls(1);
    } else {
        return false;
    }
}

} // namespace starrocks::vectorized::csv
